#ifndef __ANALYSISREPORT_H__
#define __ANALYSISREPORT_H__

#include <map>
#include <set>
#include "Report.h"
#include "Declaration.h"

class AnalysisReport : public Report
{
	private:
		typedef std::multimap<Declaration,Declaration,Declaration::Compare> DDMMap;
		typedef std::multimap<int,std::string,std::less<int> > ISMMap;
		typedef std::multimap<std::string,std::string,std::less<std::string> > SSMMap;
		typedef std::set<Declaration,Declaration::Compare> DSet;
		typedef std::set<std::string,std::less<std::string> > SSet;
		class DataCollectorContext
		{
			private:
				int scope;
				ISMMap decls;
			private:
				DataCollectorContext(const DataCollectorContext &)
					{}
				void remove_all_above_scope_level(int level)
					{
						bool quit;
						do {
							quit = true;
							for (ISMMap::iterator it = decls.begin(); it != decls.end(); ++it)
								if (it->first >= level)
								{
									decls.erase(it->first);
									quit = false;
									break;
								}
						} while (!quit);
					}
			public:
				DataCollectorContext()
					: scope(0)
					{}
				~DataCollectorContext()
					{ decls.clear(); }
				inline void add(const std::string & decl)
					{ decls.insert(ISMMap::value_type(scope, decl)); }
				inline void enter_scope(void)
					{ ++scope; }
				inline void leave_scope(void)
					{ remove_all_above_scope_level(scope--); }
				inline int get_decl_scope(const std::string & symb) const
					{
						for (ISMMap::const_reverse_iterator r = decls.rbegin(); r != decls.rend(); ++r)
							if (r->second == symb) return r->first;
						return -1;
					}
				inline bool is_declared(const std::string & symb) const
					{ return (get_decl_scope(symb) >= 0); }
		};
	private:
		SSet modules;		// set: modules
		SSMMap functions;	// map: module => function(s)
		DSet funcs;			// set: all functions
		DDMMap callgraph;	// map: caller => callee(s)
		DSet sdata;			// set: all static data
		DSet gdata;			// set: all global data
		DDMMap data_access;	// data access from functions (map: function -> [sg]data)
		Declaration caller;	// temporary object
	private:
		bool is_function_defined_in_module(const std::string &, const std::string &) const;
		std::string get_function_name(const Node *) const;
		std::string find_module_of_function(const std::string &) const;
		std::string get_identifier_of_parameter(const Node *) const;
		bool extract_identifier_of_direct_declarator(const Node *, std::string &) const;
		void extract_identifiers_from_declaration(const Node *, DataCollectorContext &);
		// collectors
		void collect_defined_functions(const STVec &);
		void collect_function_info(const STVec &);
		void collect_function_calls(Node *);
		void collect_data_from_declaration(const Node *);
		void collect_data(const STVec &);
		void collect_function_parameters(const Node *, DataCollectorContext &);
		void collect_data_recursive_depth(const Node *, DataCollectorContext &);
		void collect_data_accesses_from_func(const Node *);
		void collect_unique_data_access(const Declaration &);
		// output
		void print_defined_functions(std::ostream &) const;
		void print_calls(std::ostream &) const;
		void print_static_data(std::ostream &) const;
		void print_global_data(std::ostream &) const;
		void print_data_accesses(std::ostream &) const;
		void print_modules(std::ostream &) const;
	public:
		virtual ~AnalysisReport();
		virtual std::string name() const;
		virtual int print(const STVec &, std::ostream &);
};

#endif
