// analysis.cpp

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <set>
#include <map>
#include <vector>
#include <getopt.h>

using namespace std;

#include "Declaration.h"
#include "stringutils.h"
#include "version.h"

typedef multimap<Declaration,Declaration,Declaration::Compare> DDMMap;
typedef multimap<string,Declaration,less<string> > SDMMap;
typedef multimap<string,string,less<string> > SSMMap;
typedef map<Declaration,string,Declaration::Compare> DSMap;
typedef map<string,Declaration,less<string> > SDMap;
typedef map<string,string,less<string> > SSMap;
typedef set<Declaration,Declaration::Compare> DSet;
typedef set<string,less<string> > SSet;

string prog_name = "";
string data_filename = "";
string graph_parameter = "";
string subgraph_parameter = "";

// data is highly redundant, but easier to browse
SSet modules;			// list of modules
DSet functions;			// list of functions
DDMMap ct;				// calltree (calling function -> called function)
DDMMap ict;				// inverse calltree (called function -> calling function)
DSet sdata;				// static data
DSet gdata;				// global data
DDMMap fad;				// function accessing data map (function -> data)
DDMMap daf;				// data accessed from function map (data -> function)
SDMMap mod_to_func;		// mapping of (module -> functions)
SDMMap mod_to_static;	// mapping of (module -> static data)
SDMMap mod_to_global;	// mapping of (module -> global data)

// graphviz needs ID (charset [_0-9a-zA-Z]) for nodes,
// real names are to be done using attribute 'label' for nodes
DSMap fid;				// unique function IDs  (func -> id)
SDMap idf;				// unique function IDs  (id -> func)
SSMap mid;				// unique module IDs    (module -> id)
SSMap idm;				// unique module IDs    (id -> module)
DSMap sdid;				// unique static data IDs (static data -> id)
SDMap idsd;				// unique static data IDs (id -> static data)
DSMap gdid;				// unique global data IDs (global data -> id)
SDMap idgd;				// unique global data IDs (id -> global data)

class IDGenerator
{
	private:
		int num;
		std::string prefix;
		int num_width;
	public:
		IDGenerator(const std::string prefix, int num_width = 4)
			: num(0), prefix(prefix), num_width(num_width)
			{}
		inline std::string get(void)
			{
				ostringstream oss(ostringstream::out);
				oss << prefix << setfill('0') << setw(num_width) << num++;
				return oss.str();
			}
};

void read_data(void)
{
	// open file
	static const string default_data_filename = "analysis.dat";
	string filename = (data_filename.size() > 0) ? data_filename : default_data_filename;
	ifstream ifs(filename.c_str(), ios::in);
	if (!ifs)
	{
		cerr << "cannot open file '" << filename << "'" << endl;
		exit(-1);
	}
	// read file
	string line;
	//size_t delm;
	char type;
	while (getline(ifs, line))
	{
		trim_in_place(line);
		if (line.empty()) continue;
		istringstream is(line);
		is >> type;
		switch (type)
		{
			case 'M': { // module
				string mod;
				is >> mod;
				modules.insert(mod);
				} break;
			case 'F': { // function
				Declaration func;
				is >> func;
				modules.insert(func.get_module());
				functions.insert(func);
				mod_to_func.insert(SDMMap::value_type(func.get_module(), func));
				} break;
			case 'C': { // call
				Declaration caller, callee;
				is >> caller >> callee;
				ct.insert(DDMMap::value_type(caller, callee));
				ict.insert(DDMMap::value_type(callee, caller));
				} break;
			case 'S': { // static data
				Declaration decl;
				is >> decl;
				sdata.insert(decl);
				mod_to_static.insert(SDMMap::value_type(decl.get_module(), decl));
				} break;
			case 'G': { // global data
				Declaration decl;
				is >> decl;
				gdata.insert(decl);
				mod_to_global.insert(SDMMap::value_type(decl.get_module(), decl));
				} break;
			case 'A': { // data accesses
				Declaration func;
				Declaration data;
				is >> func >> data;
				fad.insert(DDMMap::value_type(func, data));
				daf.insert(DDMMap::value_type(data, func));
				} break;
			default:
				cerr << "unknown entry type '" << type << "' in file '" << filename << "'" << endl;
				exit(-1);
		}
	}
	// create IDs for functions
	IDGenerator func_id("func_");
	for (DSet::const_iterator i = functions.begin(); i != functions.end(); ++i)
	{
		string id = func_id.get();
		fid.insert(DSMap::value_type(*i, id));
		idf.insert(SDMap::value_type(id, *i));
	}
	// create IDs for modules
	IDGenerator mod_id("module_");
	for (SSet::const_iterator i = modules.begin(); i != modules.end(); ++i)
	{
		string id = mod_id.get();
		mid.insert(SSMap::value_type(*i, id));
		idm.insert(SSMap::value_type(id, *i));
	}
	// create IDs for static data
	IDGenerator sd_id("static_");
	for (DSet::const_iterator i = sdata.begin(); i != sdata.end(); ++i)
	{
		string id = sd_id.get();
		sdid.insert(DSMap::value_type(*i, id));
		idsd.insert(SDMap::value_type(id, *i));
	}
	// create IDs for global data
	IDGenerator gd_id("global_");
	for (DSet::const_iterator i = gdata.begin(); i != gdata.end(); ++i)
	{
		string id = gd_id.get();
		gdid.insert(DSMap::value_type(*i, id));
		idgd.insert(SDMap::value_type(id, *i));
	}
}

void graphviz_module_frame(ostream & os, DSet & visited)
{
	SDMMap mod_func_map;
	SSet mods;
	for (DSet::const_iterator it = visited.begin(); it != visited.end(); ++it)
	{
		mods.insert(it->get_module());
		mod_func_map.insert(SDMap::value_type(it->get_module(), *it));
	}
	size_t cluster = 0;
	for (SSet::const_iterator it = mods.begin(); it != mods.end(); ++it)
	{
		os << "\tsubgraph cluster" << cluster++ << " {" << endl;
		os << "\t\tlabel=\"" << replace_const(*it, "\\", "/") << "\";" << endl;
		os << subgraph_parameter << endl;
		for (SDMMap::const_iterator i = mod_func_map.lower_bound(*it); i != mod_func_map.upper_bound(*it); ++i)
			os << "\t\t" << fid[i->second] << " [shape=ellipse,label=\"" << i->second.get_name() << "\"];" << endl;
		os << "\t}" << endl;
	}
}

void graphviz_calltree(ostream & os, DDMMap * container, const Declaration & func, int depth, DSet & visited, bool forward_arrows = true)
{
	if (depth == 0)
	{
		visited.insert(func);
		return;
	}
	if (visited.find(func) != visited.end()) return; // already visited?
	for (DDMMap::const_iterator it = container->lower_bound(func); it != container->upper_bound(func); ++it)
	{
		if (forward_arrows) os << "\t" << fid[it->first] << " -> " << fid[it->second] << " ;" << endl;
		else os << "\t" << fid[it->second] << " -> " << fid[it->first] << " ;" << endl;
	}
	visited.insert(func);
	for (DDMMap::const_iterator it = container->lower_bound(func); it != container->upper_bound(func); ++it)
		graphviz_calltree(os, container, it->second, (depth > 0) ? depth-1 : depth, visited, forward_arrows);
}

void graphviz_calltree(ostream & os, const string & mod_func, int depth = -1)
{
	Declaration func;
	if (Declaration::parse(mod_func, func) == false)
	{
		cerr << prog_name << ": function not in the form \"module:function\"" << endl;
		exit(-1);
	}
	if (depth < -1) depth = -1;
	DSet visited;
	graphviz_calltree(os, &ct, func, depth, visited, true);
	graphviz_module_frame(os, visited);
}

void graphviz_inv_calltree(ostream & os, const string & mod_func, int depth = -1)
{
	Declaration func;
	if (Declaration::parse(mod_func, func) == false)
	{
		cerr << prog_name << ": function not in the form \"module:function\"" << endl;
		exit(-1);
	}
	if (depth < -1) depth = -1;
	DSet visited;
	graphviz_calltree(os, &ict, func, depth, visited, false);
	graphviz_module_frame(os, visited);
}

void collect_module_dependencies(const string & module, SSet & required_modules, SSet & calling_modules)
{
	for (SDMMap::const_iterator it = mod_to_func.lower_bound(module); it != mod_to_func.upper_bound(module); ++it)
	{
		// all required modules (modules containing functions that are called from function in the specified module)
		for (DDMMap::const_iterator i = ct.lower_bound(it->second); i != ct.upper_bound(it->second); ++i)
		{
			if (i->second.get_module() == module) continue; // dependencies within the module are not shown
			if (i->second.get_module().empty()) continue;
			required_modules.insert(i->second.get_module());
		}
		// all calling modules (modules containing functions that are calling functions of the spefified module)
		for (DDMMap::const_iterator i = ict.lower_bound(it->second); i != ict.upper_bound(it->second); ++i)
		{
			if (i->second.get_module() == module) continue; // dependencies within the module are not shown
			if (i->second.get_module().empty()) continue; // neccessary because TOP has no caller
			calling_modules.insert(i->second.get_module());
		}
	}
}

void graphviz_module_dep(ostream & os, const string & module)
{
	SSet required_modules;
	SSet calling_modules;
	// collect data
	collect_module_dependencies(module, required_modules, calling_modules);
	string module_id = mid[module];
	// describe graph
	os << "\t" << module_id << " [shape=box,label=\"" << replace_const(module, "\\", "/") << "\"];" << endl;
	for (SSet::const_iterator it = required_modules.begin(); it != required_modules.end(); ++it)
	{
		string id = mid[*it];
		os << "\t" << module_id << "  ->  " << id << " ;" << endl;
		os << "\t" << id << " [shape=box,label=\"" << replace_const(*it, "\\", "/") << "\"];" << endl;
	}
	for (SSet::const_iterator it = calling_modules.begin(); it != calling_modules.end(); ++it)
	{
		string id = mid[*it];
		os << "\t" << id << "  ->  " << module_id << " ;" << endl;
		os << "\t" << id << " [shape=box,label=\"" << replace_const(*it, "\\", "/") << "\"];" << endl;
	}
}

void graphviz_all_module_deps(ostream & os)
{
	SSMMap mmdep; // (module ID)->(module ID) dependency
	// collect data
	for (SSet::const_iterator module = modules.begin(); module != modules.end(); ++module)
	{
		string module_id = mid[*module];
		for (SDMMap::const_iterator it = mod_to_func.lower_bound(*module); it != mod_to_func.upper_bound(*module); ++it)
		{
			for (DDMMap::const_iterator i = ct.lower_bound(it->second); i != ct.upper_bound(it->second); ++i)
			{
				if (*module == i->second.get_module()) continue; // dependencies within the module are not shown
				if (i->second.get_module().empty()) continue;
				bool found = false;
				for (SSMMap::const_iterator f = mmdep.lower_bound(*module); !found && f != mmdep.upper_bound(*module); ++f)
					found = (f->second == i->second.get_module());
				if (!found) mmdep.insert(SSMMap::value_type(module_id, mid[i->second.get_module()]));
			}
		}
	}
	// describe graph
	for (SSet::const_iterator it = modules.begin(); it != modules.end(); ++it)
		os << "\t" << mid[*it] << " [shape=box,label=\"" << replace_const(*it, "\\", "/") << "\"];" << endl;
	for (SSMMap::const_iterator it = mmdep.begin(); it != mmdep.end(); ++it)
		os << "\t" << it->first << "  ->  " << it->second << " ;" << endl;
}

void graphviz_module_members(ostream & os, const string & module)
{
	// output: defined functions
	for (SDMMap::const_iterator i = mod_to_func.lower_bound(module); i != mod_to_func.upper_bound(module); ++i)
		os << "\t\t" << fid[i->second] << " [shape=ellipse,label=\"" << i->second.get_name() << "\"];" << endl;
	// output: static data
	for (SDMMap::const_iterator i = mod_to_static.lower_bound(module); i != mod_to_static.upper_bound(module); ++i)
		os << "\t\t" << sdid[i->second] << " [shape=box,label=\"" << i->second.get_name() << "\"];" << endl;
	// output: access of data from functions, only within this module and function calls within the module
	for (SDMMap::const_iterator i = mod_to_func.lower_bound(module); i != mod_to_func.upper_bound(module); ++i)
	{
		string func_module = i->second.get_module();
		string func_id = fid[i->second];
		for (DDMMap::const_iterator j = fad.lower_bound(i->second); j != fad.upper_bound(i->second); ++j)
			if (func_module == j->second.get_module())
			{
				if (sdid.find(j->second) == sdid.end()) continue; // only static data, no global data
				os << "\t\t" << func_id << " -> " << sdid[j->second] << " ;" << endl;
			}
		for (DDMMap::const_iterator j = ct.lower_bound(i->second); j != ct.upper_bound(i->second); ++j)
			if (func_module == j->second.get_module())
				os << "\t\t" << func_id << " -> " << fid[j->second] << " ;" << endl;			
	}
}

void graphviz_subgraph_module_members(ostream & os, int subgraph_id, const string & module)
{
	os << "\tsubgraph cluster" << subgraph_id << " {" << endl;
	os << "\t\tlabel=\"" << replace_const(module, "\\", "/") << "\";" << endl;
	os << subgraph_parameter << endl;
	graphviz_module_members(os, module);
	os << "\t}" << endl;
}

void graphviz_all_module_members(ostream & os, const string & module)
{
	graphviz_subgraph_module_members(os, 0, module);
}

void graphviz_global_access(ostream & os, const string & data_str, int depth)
{
	// TODO: (functioncall-)depth is not considered yet!
	Declaration data;
	if (!Declaration::parse(data_str, data)) return;
	SSet mods;
	DSet funcs;
	for (DDMMap::const_iterator i = daf.lower_bound(data); i != daf.upper_bound(data); ++i)
	{
		os << "\t" << fid[i->second] << " -> " << gdid[i->first] << " ;" << endl;
		funcs.insert(i->second);
		mods.insert(i->second.get_module());
	}
	bool mod_for_global_drawn = false;
	int cluster = 0;
	for (SSet::const_iterator i = mods.begin(); i != mods.end(); ++i)
	{
		os << "\t" << "subgraph cluster" << cluster++ << " {" << endl;
		os << "\t\tlabel=\"" << replace_const(*i, "\\", "/") << "\";" << endl;
		os << subgraph_parameter << endl;
		if (*i == data.get_module())
		{
			os << "\t\t" << gdid[data] << " [shape=box,label=\"" << data.get_name() << "\"];" << endl;
			mod_for_global_drawn = true;
		}
		for (DSet::const_iterator j = funcs.begin(); j != funcs.end(); ++j)
			if (j->get_module() == *i)
				os << "\t\t" << fid[*j] << " [shape=ellipse,label=\"" << j->get_name() << "\"];" << endl;
		os << "\t" << "}" << endl;
	}
	if (!mod_for_global_drawn)
	{
		os << "\t" << "subgraph cluster" << cluster++ << " {" << endl;
		os << "\t\tlabel=\"" << replace_const(data.get_module(), "\\", "/") << "\";" << endl;
		os << subgraph_parameter << endl;
		os << "\t\t" << gdid[data] << " [shape=box,label=\"" << data.get_name() << "\"];" << endl;
		os << "\t" << "}" << endl;
	}
}

void graphviz_global_all_access(ostream & os, int depth)
{
	// TODO: (functioncall-)depth is not considered yet!
	SSet gdmods;
	SSet fmods;
	DSet fadfuncs;
	for (DSet::const_iterator i = gdata.begin(); i != gdata.end(); ++i) gdmods.insert(i->get_module());
	// draw all acesses
	for (DDMMap::const_iterator i = fad.begin(); i != fad.end(); ++i)
	{
		if (gdata.find(i->second) != gdata.end())
		{
			os << "\t" << fid[i->first] << " -> " << gdid[i->second] << " ;" << endl;
			fmods.insert(i->first.get_module());
			fadfuncs.insert(i->first);
		}
	}
	// draw all subgraphs
	int cluster = 0;
	for (SSet::const_iterator i = fmods.begin(); i != fmods.end(); ++i)
	{
		os << "\t" << "subgraph cluster" << cluster++ << " {" << endl;
		os << "\t\tlabel=\"" << replace_const(*i, "\\", "/") << "\";" << endl;
		os << subgraph_parameter << endl;
		// print functions defined in module
		for (SDMMap::const_iterator j = mod_to_func.lower_bound(*i); j != mod_to_func.upper_bound(*i); ++j)
			if (fadfuncs.find(j->second) != fadfuncs.end())
				os << "\t\t" << fid[j->second] << " [shape=ellipse,label=\"" << j->second.get_name() << "\"];" << endl;
		// print globals defined in module
		for (SDMMap::const_iterator j = mod_to_global.lower_bound(*i); j != mod_to_global.upper_bound(*i); ++j)
			os << "\t\t" << gdid[j->second] << " [shape=box,label=\"" << j->second.get_name() << "\"];" << endl;
		// TODO: function calls, only between functions of 'fadfuncs'
		os << "\t" << "}" << endl;
		gdmods.erase(*i);
	}
	for (SSet::const_iterator i = gdmods.begin(); i != gdmods.end(); ++i)
	{
		os << "\t" << "subgraph cluster" << cluster++ << " {" << endl;
		os << "\t\tlabel=\"" << replace_const(*i, "\\", "/") << "\";" << endl;
		os << subgraph_parameter << endl;
		// print globals defined in module
		for (SDMMap::const_iterator j = mod_to_global.lower_bound(*i); j != mod_to_global.upper_bound(*i); ++j)
			os << "\t\t" << gdid[j->second] << " [shape=box,label=\"" << j->second.get_name() << "\"];" << endl;
		os << "\t" << "}" << endl;
		gdmods.erase(*i);
	}
}

void graphviz_module_dep_detail(ostream & os, const string & module)
{
	SSet required_modules;
	SSet calling_modules;
	// collect data
	collect_module_dependencies(module, required_modules, calling_modules);
	string module_id = mid[module];

	// describe graph
	int k = 0;
	graphviz_subgraph_module_members(os, k++, module);
	for (SSet::const_iterator it = required_modules.begin(); it != required_modules.end(); ++it)
		graphviz_subgraph_module_members(os, k++, *it);
	for (SSet::const_iterator it = calling_modules.begin(); it != calling_modules.end(); ++it)
		graphviz_subgraph_module_members(os, k++, *it);

	// calltree (depth=1) for all module functions
	for (SDMMap::const_iterator it = mod_to_func.lower_bound(module); it != mod_to_func.upper_bound(module); ++it)
	{
		DSet visited;
		graphviz_calltree(os, &ct, it->second, 1, visited, true);
	}
	// inverse calltree (depth=1) for all module functions
	for (SDMMap::const_iterator it = mod_to_func.lower_bound(module); it != mod_to_func.upper_bound(module); ++it)
	{
		DSet visited;
		graphviz_calltree(os, &ict, it->second, 1, visited, false);
	}
	// accesses of module functions to globals
	for (SDMMap::const_iterator i = mod_to_func.lower_bound(module); i != mod_to_func.upper_bound(module); ++i)
	{
		for (DDMMap::const_iterator j = fad.lower_bound(i->second); j != fad.upper_bound(i->second); ++j)
		{
			if (j->second.get_module() != j->first.get_module())
			{
				os << "\t" << gdid[j->second] << " [shape=box,label=\"" << j->second.get_name() << "\"];" << endl;
				os << "\t" << fid[i->second] << " -> " << gdid[j->second] << " ;" << endl;
			}
		}
	}
}

enum Command
{
	CMD_NONE = -1
	,CMD_UNKNOWN = 0
	,CMD_HELP
	,CMD_CALLTREE
	,CMD_INV_CALLTREE
	,CMD_GLOBAL_ACCESS
	,CMD_GLOBAL_ALL_ACCESS
	,CMD_MODULE_MEMBERS
	,CMD_MODULE_DEPENDENCIES
	,CMD_MODULE_DEPENDENCIES_DETAIL
	,CMD_ALL_MODULE_DEPENDENCIES
};

typedef struct
{
	Command cmd;
	string arg;
	int depth;
	string output;
} cmd_params_t;

static const struct option long_options[] =
{
	{ "help",  	     no_argument,       0, 'h' },
	{ "ct",    	     required_argument, 0, 'c' },
	{ "ict",   	     required_argument, 0, 'i' },
	{ "global",      required_argument, 0, 'g' },
	{ "global-all",  no_argument,       0, 'G' },
	{ "mdep",  	     required_argument, 0, 'm' },
	{ "mdep-detail", required_argument, 0, 'M' },
	{ "adep",  	     no_argument,       0, 'a' },
	{ "depth", 	     required_argument, 0, 'd' },
	{ "output",      required_argument, 0, 'o' },
	{ "members",     required_argument, 0, 'b' },
	{ 0,             0,                 0, 0   }
};
static const char * opt_string = "hc:i:m:M:ad:o:b:g:Gf:";

void usage(void)
{
	cerr << endl;
	cerr << "Analysis " << VERSION << endl;
	cerr << "usage: " << prog_name
		<< " [-f|--file datafile]"
		<< " [-c|--ct function]"
		<< " [-i|--ict fuction]"
		<< " [-g|--global global]"
		<< " [-G|--global-all]"
		<< " [-b|--members module]"
		<< " [-m|--mdep module]"
		<< " [-M|--mdep-detail module]"
		<< " [-a|--adep]"
		<< " [-d|--depth num]"
		<< " [-o|--output filename]"
		<< endl;
	cerr << "    -f,--file        : use the specified data file, instead of 'analysis.dat'" << endl;
	cerr << "    -c,--ct          : calltree for \"module:function\"" << endl;
	cerr << "    -i,--ict         : inverse calltree for \"module:function\"" << endl;
	cerr << "    -g,--global      : access diagram for specified global data \"module:data\"" << endl;
	cerr << "    -G,--global-all  : access diagram for all global data" << endl;
	cerr << "    -b,--members     : members of the spefified module" << endl;
	cerr << "    -m,--mdep        : dependencies of the specified module" << endl;
	cerr << "    -M,--mdep-detail : dependencies of the specified module with its details" << endl;
	cerr << "    -a,--adep        : dependencies of all modules" << endl;
	cerr << "    -d,--depth       : depth of normal and inverted calltrees, number > 1 (default=0: infinite)" << endl;
	cerr << "    -o,--output      : output filename, instead of stdout" << endl;
	cerr << endl;
	exit(-1);
}

void parse_parameters(int argc, char ** argv, cmd_params_t & param)
{
	param.cmd = CMD_NONE;
	param.arg = "";
	param.depth = -1;
	param.output = "";
	while (optind < argc)
	{
		int index = -1;
		switch (getopt_long(argc, argv, opt_string, long_options, &index))
		{
			case 'h': usage(); break;
			case 'c':
				param.cmd = CMD_CALLTREE;
				param.arg = optarg;
				break;
			case 'i':
				param.cmd = CMD_INV_CALLTREE;
				param.arg = optarg;
				break;
			case 'm':
				param.cmd = CMD_MODULE_DEPENDENCIES;
				param.arg = optarg;
				break;
			case 'M':
				param.cmd = CMD_MODULE_DEPENDENCIES_DETAIL;
				param.arg = optarg;
				break;
			case 'a':
				param.cmd = CMD_ALL_MODULE_DEPENDENCIES;
				break;
			case 'd':
				istringstream(optarg) >> param.depth;
				if (param.depth <= 0) param.depth = -1;
				break;
			case 'o':
				param.output = optarg;
				break;
			case 'b':
				param.cmd = CMD_MODULE_MEMBERS;
				param.arg = optarg;
				break;
			case 'g':
				param.cmd = CMD_GLOBAL_ACCESS;
				param.arg = optarg;
				break;
			case 'G':
				param.cmd = CMD_GLOBAL_ALL_ACCESS;
				break;
			case 'f':
				data_filename = optarg;
				break;
			case -1:
				param.cmd = CMD_NONE;
				break;
			default:
			case 0:
				param.cmd = CMD_UNKNOWN;
				break;
		}
	}
}

void load_graph_parameter(const string & filename, string & parameter)
{
	ifstream ifs(filename.c_str(), ios::in);
	if (ifs)
	{
		ostringstream oss;
		string line;
		while (getline(ifs, line)) oss << line << endl;
		ifs.close();
		parameter = oss.str();
	}
}

int main(int argc, char ** argv)
{
	// extract program name
	prog_name = argv[0];
	prog_name = prog_name.substr(prog_name.find_last_of("/\\")+1);

	// parse command line parameters
	cmd_params_t param;
	if (argc < 2) usage();
	parse_parameters(argc, argv, param);

	// read data and parameter
	read_data();
	load_graph_parameter("graphprm.txt", graph_parameter);
	load_graph_parameter("subgraphprm.txt", subgraph_parameter);

	// redirect cout if an output file is specified
	streambuf * cout_buffer = cout.rdbuf();
	ofstream ofs;
	if (param.output != "")
	{
		ofs.open(param.output.c_str(), ios::out | ios::trunc);
		if (!ofs)
		{
			cerr << argv[0] << ": error: cannot open output file '" << param.output << "'" << endl;
			cerr << endl;
			exit(-2);
		}
		cout.rdbuf(ofs.rdbuf());
	}

	// execute command
	switch (param.cmd)
	{
		default:
		case CMD_UNKNOWN:
			usage();
			break;
		case CMD_CALLTREE:
			cout << "digraph callgraph {" << endl;
			cout << graph_parameter << endl;
			graphviz_calltree(cout, param.arg, param.depth);
			cout << "}" << endl;
			break;
		case CMD_INV_CALLTREE:
			cout << "digraph inverted_callgraph {" << endl;
			cout << graph_parameter << endl;
			graphviz_inv_calltree(cout, param.arg, param.depth);
			cout << "}" << endl;
			break;
		case CMD_GLOBAL_ACCESS:
			cout << "digraph access_to_global {" << endl;
			cout << graph_parameter << endl;
			graphviz_global_access(cout, param.arg, param.depth);
			cout << "}" << endl;
			break;
		case CMD_GLOBAL_ALL_ACCESS:
			cout << "digraph access_to_all_globals {" << endl;
			cout << graph_parameter << endl;
			graphviz_global_all_access(cout, param.depth);
			cout << "}" << endl;
			break;
		case CMD_MODULE_MEMBERS:
			cout << "digraph module_members {" << endl;
			cout << graph_parameter << endl;
			graphviz_all_module_members(cout, param.arg);
			cout << "}" << endl;
			break;
		case CMD_MODULE_DEPENDENCIES:
			cout << "digraph module_dep {" << endl;
			cout << graph_parameter << endl;
			graphviz_module_dep(cout, param.arg);
			cout << "}" << endl;
			break;
		case CMD_MODULE_DEPENDENCIES_DETAIL:
			cout << "digraph module_dep_detail {" << endl;
			cout << graph_parameter << endl;
			graphviz_module_dep_detail(cout, param.arg);
			cout << "}" << endl;
			break;
		case CMD_ALL_MODULE_DEPENDENCIES:
			cout << "digraph all_module_dep {" << endl;
			cout << graph_parameter << endl;
			graphviz_all_module_deps(cout);
			cout << "}" << endl;
			break;
	}

	// restore cout
	cout.rdbuf(cout_buffer);

	// features:
	// - call tree (for a function), depth configurable
	// - inverse call tree (for a function), depth configurable
	// - module (file) dependencies
	// - dependencies beteween all modules
	// - global/static data

	// future features:
	// - warnings:
	//   - recursions (direct and indirect)
	//   - name clashes
	//   - circular dependencies

	return 0;
}
